@@ -3,9 +3,10 @@ from datetime import date, timedelta
 
 from dateutil.relativedelta import relativedelta
 from django.contrib.contenttypes.models import ContentType
+from django.core.exceptions import ValidationError
 from django.db import models
 from django.conf import settings
-from django.core.validators import MinValueValidator
+from django.core.validators import MinValueValidator, EMPTY_VALUES
 from django.urls import reverse
 from django.utils.timezone import now
 from django.utils.translation import ugettext_lazy as _, override as override_language
@@ -17,7 +18,7 @@ from fees.querysets import PackageQuerySet, PlanQuerySet
 from fees import settings as fees_settings
 from .helpers import get_purchaser_model
 
-logger = logging.getLogger('fee')
+logger = logging.getLogger('fees')
 
 
 class Quota(models.Model):
@@ -29,15 +30,16 @@ class Quota(models.Model):
     unit = models.CharField(_('unit'), max_length=100, blank=True)
     description = models.TextField(_('description'), blank=True)
     is_boolean = models.BooleanField(_('is boolean'), default=False)
+    order = models.PositiveSmallIntegerField(verbose_name=_('ordering'), help_text=_('to set order in pricing'), default=1)
     i18n = TranslationField(fields=('name', 'description'))
 
     class Meta:
-        ordering = ('name',)  # TODO: custom ordering
+        ordering = ('order',)
         verbose_name = _("Quota")
         verbose_name_plural = _("Quotas")
 
     def __str__(self):
-        return "%s" % (self.codename, )
+        return "%s" % (self.name_i18n, )
 
 
 class PackageQuotaManager(models.Manager):
@@ -54,6 +56,7 @@ class PackageQuota(models.Model):
     class Meta:
         verbose_name = _("Package quota")
         verbose_name_plural = _("Packages quotas")
+        unique_together = (('package', 'quota'),)
 
 
 class Package(models.Model):
@@ -62,7 +65,14 @@ class Package(models.Model):
     order = models.PositiveSmallIntegerField(verbose_name=_('ordering'), help_text=_('to set order in pricing'), unique=True, default=1)
     trial_duration = models.PositiveSmallIntegerField(verbose_name=_('trial duration'), help_text=_('in days'), default=0)
     is_default = models.BooleanField(
-        help_text=_('Default package for purchaser'),
+        _('default'),
+        help_text=_('Default package for new purchaser (useful for trial packages). Only 1 default package at a time can be set.'),
+        default=False,
+        db_index=True,
+    )
+    is_fallback = models.BooleanField(
+        _('fallback'),
+        help_text=_('Fallback package for purchaser when its subscription expires or trial ends (useful for freemium packages). Only 1 fallback package at a time can be set.'),
         default=False,
         db_index=True,
     )
@@ -78,15 +88,19 @@ class Package(models.Model):
     created = models.DateTimeField(_('created'), auto_now_add=True)
     modified = models.DateTimeField(_('modified'), auto_now=True)
     objects = PackageQuerySet.as_manager()
-    i18n = TranslationField(fields=('description',))
+    i18n = TranslationField(fields=('title', 'description',))
 
     class Meta:
         verbose_name = _('package')
         verbose_name_plural = _('packages')
         ordering = ['order']
+        constraints = [
+            models.UniqueConstraint(fields=['is_default'], condition=models.Q(is_default=True), name='unique_default'),
+            models.UniqueConstraint(fields=['is_fallback'], condition=models.Q(is_fallback=True), name='unique_fallback'),
+        ]
 
     def __str__(self):
-        return self.title
+        return self.title_i18n
 
     def get_absolute_url(self):
         return reverse('fees:packages')
@@ -99,32 +113,70 @@ class Package(models.Model):
             return_value = None
         return return_value
 
+    @classmethod
+    def get_fallback_package(cls):
+        try:
+            return_value = cls.objects.get(is_fallback=True)
+        except cls.DoesNotExist:
+            return_value = None
+        return return_value
+
     @classmethod
     def get_current_package(cls, purchaser):
-        """ Get current package for purchaser. If plan is expired, return None """
-        # We need to handle both default package (new purchaser -> TRIAL) and expired plan -> None
+        """
+        Get current package for purchaser.
+        App should be responsible for creating purchaser's plan (after sign up for example).
+        If plan is expired or not present, return fallback package if available else None.
+        """
+        # We need to handle both default package (new purchaser -> TRIAL) and expired plan -> fallback
 
         is_anonymous = isinstance(purchaser, get_user_model()) and purchaser.is_anonymous
 
-        if not purchaser or is_anonymous or not hasattr(purchaser, 'plan') or purchaser.plan.is_expired():
+        # anonymous user
+        if not purchaser or is_anonymous:
             # TODO: any other rules?
             return None
-            # default_package = Package.get_default_package()
-            # if default_package is None or not default_package.is_free():
-            #     raise ValidationError(_('Plan has expired'))
-            # return default_plan
-        return purchaser.plan.package
+
+        plan_is_set = hasattr(purchaser, 'plan') and purchaser.plan is not None
+
+        # currently, valid (not expired) package
+        if plan_is_set and not purchaser.plan.is_expired():
+            return purchaser.plan.package
+
+        # fallback package
+        fallback_package = Package.get_fallback_package()
+
+        # validation of fallback package price (fallback package can't be free)
+        if fallback_package is not None and not fallback_package.is_free():
+            if plan_is_set:
+                raise ValidationError(_('Plan has expired and fallback package is not free'))
+            else:
+                raise ValidationError(_('Plan not found and fallback package is not free'))
+
+        # fallback package for expired or not present plan
+        return fallback_package
 
     def get_quotas(self):
         quota_dic = {}
-        for plan_quota in PackageQuota.objects.filter(package=self).select_related('quota'):
-            quota_dic[plan_quota.quota.codename] = plan_quota.value
+        for package_quota in PackageQuota.objects.filter(package=self).select_related('quota'):
+            if Package.is_quota_available(package_quota):
+                quota_dic[package_quota.quota.codename] = package_quota.value
         return quota_dic
 
+    @staticmethod
+    def is_quota_available(package_quota):
+        value = package_quota['value'] if isinstance(package_quota, dict) else package_quota.value
+        return value is None or value > 0
+
     def is_free(self):
         return not self.pricing_set.exists()
     is_free.boolean = True
 
+    def monthly_plan(self):
+        return self.pricing_set.filter(period=Pricing.PERIOD_MONTH).order_by('duration').first()
+
+    is_free.boolean = True
+
 
 class Pricing(models.Model):
     PERIOD_DAY = 'DAY'
@@ -153,12 +205,31 @@ class Pricing(models.Model):
     # )
     created = models.DateTimeField(_('created'), auto_now_add=True)
     modified = models.DateTimeField(_('modified'), auto_now=True)
-    objects = PackageQuerySet.as_manager()
+    # objects = PricingQuerySet.as_manager()
 
     class Meta:
         verbose_name = _('pricing')
-        verbose_name_plural = _('pricing plans')
+        verbose_name_plural = _('pricings')
         ordering = ['price']
+        unique_together = (['package', 'period', 'duration'],)
+
+    @classmethod
+    def distinct_period_durations(cls):
+        period_durations = cls.objects.order_by('period', 'duration').values('period', 'duration').distinct()
+
+        for period_duration in period_durations:
+            period_duration['display'] = Pricing.period_duration_display(
+                period=period_duration['period'],
+                duration=period_duration['duration']
+            )
+
+        return period_durations
+
+    @staticmethod
+    def period_duration_display(period, duration):
+        period_localize = dict(Pricing.PERIODS_PLURALIZE).get(period)
+        period_display = period_localize[0] if duration == 1 else period_localize[1]  # TODO: i18n
+        return f'{duration} {period_display}'
 
     def __str__(self):
         return f'{self.package} ({self.get_duration_display()})'
@@ -167,21 +238,19 @@ class Pricing(models.Model):
         return reverse('fees:pricing')
 
     def get_duration_display(self):
-        period_localize = dict(self.PERIODS_PLURALIZE).get(self.period)
-        period_display = period_localize[0] if self.duration == 1 else period_localize[1]  # TODO: i18n
-        return f'{self.duration} {period_display}'
+        return Pricing.period_duration_display(self.period, self.duration)
 
-    # TODO: remove commerce
     def get_price_display(self):
         return f'{self.price} {fees_settings.CURRENCY}'
 
     # TODO: remove commerce
-    def get_add_to_cart_url(self):  # TODO: move to ProductMixin
+    def get_add_to_cart_url(self):  # TODO: move to ProductMixin, template tag?
         if 'commerce' in settings.INSTALLED_APPS:
             content_type = ContentType.objects.get_for_model(self)
-            return reverse('commerce:add_to_cart', args=(content_type.id, self.id))
+            return reverse('commerce:add_to_cart', args=(content_type.id, self.id))  # TODO: human readable URL
         return '#'
-    # TODO: remove commerce
+
+    # TODO: remove commerce, add product availability handler to commerce
     # @property
     # def availability(self):
     #     from commerce.models import AbstractProduct
@@ -222,29 +291,39 @@ class Plan(models.Model):
     """
     Currently selected plan for purchaser.
     """
-    purchaser = models.OneToOneField(
+    purchaser = models.ForeignKey(
         get_purchaser_model(), verbose_name=_('purchaser'),
-        on_delete=models.CASCADE, related_name='plan', related_query_name='plan'
+        on_delete=models.CASCADE,
     )
     package = models.ForeignKey(Package, verbose_name=_('package'), on_delete=models.CASCADE)
     pricing = models.ForeignKey(Pricing, help_text=_('pricing'), default=None,
                                 null=True, blank=True, on_delete=models.CASCADE)
-    expiration = models.DateField(_('expires'), default=None, blank=True, null=True, db_index=True)
+    activation = models.DateField(_('activation'), auto_now_add=True)
+    expiration = models.DateField(_('expires'), help_text=_('keep blank to calculate automatically by pricing period'),
+        default=None, blank=True, null=True, db_index=True)
     # is_active = models.BooleanField(_('active'), default=True, db_index=True)
     # is_recurring = models.BooleanField(_('active'), default=True, db_index=True)  # TODO: can be turned on/turned off
     modified = models.DateTimeField(_('modified'), auto_now=True)
-
     objects = PlanQuerySet.as_manager()
 
     class Meta:
         verbose_name = _("Plan")
         verbose_name_plural = _("Plans")
+        ordering = ['-activation']
+        get_latest_by = 'expiration'
 
     def __str__(self):
         if self.pricing:
             return "%s [%s] (%s)" % (self.purchaser, self.package, self.pricing)
         return "%s [%s]" % (self.purchaser, self.package)
 
+    def save(self, **kwargs):
+        # update expiration date by pricing period
+        if self.expiration in EMPTY_VALUES and self.pricing:
+            from_date = self.activation or now().date()  # TODO: now().date() vs date.today()
+            self.expiration = from_date + self.pricing.timedelta
+        return super().save(**kwargs)
+
     # def is_active(self):
     #     return self.active
 
@@ -254,12 +333,17 @@ class Plan(models.Model):
         else:
             return self.expiration < date.today()
 
+    @property
     def days_left(self):
         if self.expiration is None:
             return None
         else:
             return (self.expiration - date.today()).days
 
+    def expires_soon(self):
+        days_left = self.days_left
+        return days_left is not None and days_left <= 7  # TODO: add to settings
+
     # def clean_activation(self):
     #     errors = plan_validation(self.purchaser)
     #     if not errors['required_to_activate']:
@@ -292,23 +376,23 @@ class Plan(models.Model):
     #                 days=getattr(settings, 'PLANS_DEFAULT_GRACE_PERIOD', 30))
     #         self.activate()  # this will call self.save()
 
-    def get_plan_extended_from(self, plan):
-        if plan.is_free():
+    def get_extended_from(self, package):
+        if package.is_free():
             return None
-        if not self.is_expired() and self.expiration is not None and self.plan == plan:
+        if not self.is_expired() and self.expiration is not None and self.package == package:
             return self.expiration
         return date.today()
 
-    def get_plan_extended_until(self, plan, pricing):
-        if plan.is_free():
+    def get_extended_until(self, package, pricing):
+        if package.is_free():
             return None
-        if not self.plan.is_free() and self.expiration is None:
-            return None
-        if pricing is None:
-            return self.expiration
-        return self.get_plan_extended_from(plan) + pricing.timedelta
-        # from_date = self.get_plan_extended_from(plan)
-        # return from_date + self.plan.timedelta
+        # if not self.package.is_free() and self.expiration is None:
+        #     return None  # TODO: why? which use case?
+        # if pricing is None:  # TODO: why? which use case?
+        #     return self.expiration
+        # return self.get_extended_from(package) + pricing.timedelta
+        from_date = self.get_extended_from(package)
+        return from_date + pricing.timedelta
 
     def plan_autorenew_at(self):
         """
@@ -350,7 +434,7 @@ class Plan(models.Model):
             raise ValueError(f'Extending by package {package} by invalid pricing {pricing}!')
 
         status = False  # flag; if extending account was successful?
-        new_expiration = self.get_plan_extended_until(package, pricing)
+        new_expiration = self.get_extended_until(package, pricing)
 
         if pricing is None:
             # Process a plan change request (downgrade or upgrade)
@@ -397,7 +481,7 @@ class Plan(models.Model):
             if status:
                 self.expiration = new_expiration
                 self.save()
-                logger.info("Purchaer '%s' [id=%d] has been extended by %d days using package '%s' [id=%d]" % (
+                logger.info("Purchaser '%s' [id=%d] has been extended by %d days using package '%s' [id=%d]" % (
                     self.purchaser, self.purchaser.pk, pricing.timedelta.days, package, package.pk))
                 # if getattr(settings, 'PLANS_SEND_EMAILS_PLAN_EXTENDED', True):
                 # mail_context = {'purchaser': self.purchaser,
@@ -442,7 +526,7 @@ class Plan(models.Model):
     #     mail_context = {
     #         'purchaser': self.purchaser,
     #         'plan': self,
-    #         'days': self.days_left()
+    #         'days': self.days_left
     #     }
     #     send_template_email([self.user.email], 'mail/remind_expire_title.txt', 'mail/remind_expire_body.txt',
     #                         mail_context, get_user_language(self.user))
@@ -470,8 +554,8 @@ class Plan(models.Model):
         )
 
     @classmethod
-    def create_for_users_without_plan(cls):
-        purchasers = get_purchaser_model().objects.filter(plan=None)
+    def create_for_purchasers_without_plan(cls):
+        purchasers = get_purchaser_model().objects.filter(plan=None).only('pk')
 
         for purchaser in purchasers:
             Plan.create_for_purchaser(purchaser)
