@@ -1,6 +1,7 @@
 from datetime import timedelta
 
 from django.db import models
+from django.db.models import Q
 from django.utils.timezone import now
 
 
@@ -10,6 +11,12 @@ class PackageQuerySet(models.QuerySet):
 
 
 class PlanQuerySet(models.QuerySet):
+    def active(self):
+        return self.filter(
+            Q(activation__lte=now()),
+            Q(expiration=None) | Q(expiration__gte=now()),
+        )
+
     def expires_in(self, days=7):
         threshold = now() + timedelta(days=days)
         return self.filter(expiration=threshold.date())
